unit IniFile;
// Description: INI File
// By Sarah Dean
//
// Defines INI file contents and methods for reading and writing them


interface

uses inifiles,
     dialogs,
     forms,
     classes,
     sysutils,
     Windows, // needed for WritePrivateProfileString
     IniFileConsts;

type
  TRNGType = (rngMouse, rngMicrosoft);

  TPasswordGenIniFile = class(TINIFile)
  private
    function ReadIniFileOptions() : boolean;

    function WriteIniFileOptions() : boolean;
  public
    NewVersion: boolean; // Set if this if there is no INI file or if it's for a previous version


    // Options
    OptCopyPasswordsToClipboard: boolean;
    OptWarnPasswordsCopiedToClipboard: boolean;
    OptNoStartEndInSpace: boolean;


    // Character sets available
    OptUseUppercase: boolean;
    OptUseLowercase: boolean;
    OptUseNumbers: boolean;
    OptUsePunctuation: boolean;
    OptUseSpace: boolean;

    // Punctuation set
    OptPunctuationSet: string;

    // Constraint
    OptForceCharFromEachSelectedCharSet: boolean;
    OptForceCharFromEachSelectedCharSetExceptSpace: boolean;


    // RNG
    OptRNGType: TRNGType;

    
    // Generate
    OptPasswordLength: cardinal;
    OptPasswordStrength: cardinal;
    OptTargetPasswordByLength: boolean;



    constructor Create(const FileName : string);
    destructor  Destroy(); override;

    function ReadIniFile() : boolean;
    function WriteIniFile() : boolean;

    procedure CheckVersionID();
  end;

implementation

uses
  AppGlobalConsts,
  SDUGeneral;

constructor TPasswordGenIniFile.Create(const FileName : string);
begin
  inherited;

end;

destructor TPasswordGenIniFile.Destroy();
begin

  inherited;
end;

function TPasswordGenIniFile.ReadIniFile() : boolean;
var
  retVal : boolean;
begin
  retVal := ReadIniFileOptions();

  Result := retVal;
end;


function TPasswordGenIniFile.WriteIniFile() : boolean;
var
  retVal : boolean;
begin
  retVal := WriteIniFileOptions();

  // In order to ensure that the INI file is flushed and the changes are made
  // permanent, call the Windows API function WriteProfileString() or
  // WritePrivateProfileString() and pass nil parameters for the
  // section, key, and string values. [from Borland FAQ]
  UpdateFile(); // Calls WritePrivateProfileString()

  Result := retVal;
end;


function TPasswordGenIniFile.ReadIniFileOptions() : boolean;
begin
  // Options
  OptCopyPasswordsToClipboard := ReadBool(SECTION_OPTIONS, KEY_COPY_PASSWORDS_TO_CLIPBOARD, DFLT_KEY_COPY_PASSWORDS_TO_CLIPBOARD);
  OptWarnPasswordsCopiedToClipboard := ReadBool(SECTION_OPTIONS, KEY_WARN_PASSWORDS_COPIED_TO_CLIPBOARD, DFLT_KEY_WARN_PASSWORDS_COPIED_TO_CLIPBOARD);
  OptNoStartEndInSpace := ReadBool(SECTION_OPTIONS, KEY_NO_START_END_IN_SPACE, DFLT_KEY_NO_START_END_IN_SPACE);


  // Character sets available
  OptUseUppercase   := ReadBool(SECTION_CHARACTERSETS, KEY_USE_UPPERCASE,   DFLT_KEY_USE_UPPERCASE);
  OptUseLowercase   := ReadBool(SECTION_CHARACTERSETS, KEY_USE_LOWERCASE,   DFLT_KEY_USE_LOWERCASE);
  OptUseNumbers     := ReadBool(SECTION_CHARACTERSETS, KEY_USE_NUMBERS,     DFLT_KEY_USE_NUMBERS);
  OptUsePunctuation := ReadBool(SECTION_CHARACTERSETS, KEY_USE_PUNCTUATION, DFLT_KEY_USE_PUNCTUATION);
  OptUseSpace       := ReadBool(SECTION_CHARACTERSETS, KEY_USE_SPACE,       DFLT_KEY_USE_SPACE);

  OptPunctuationSet := ReadString(SECTION_CHARACTERSETS, KEY_PUNCTUATION_SET, DFLT_KEY_PUNCTUATION_SET);
  
  OptForceCharFromEachSelectedCharSet            := ReadBool(SECTION_CHARACTERSETS, KEY_FORCE_CHAR_FROM_SELECTED_CHARSETS, DFLT_KEY_FORCE_CHAR_FROM_SELECTED_CHARSETS);
  OptForceCharFromEachSelectedCharSetExceptSpace := ReadBool(SECTION_CHARACTERSETS, KEY_FORCE_CHAR_FROM_SELECTED_CHARSETS_NOT_SPACE, DFLT_KEY_FORCE_CHAR_FROM_SELECTED_CHARSETS_NOT_SPACE);


  // RNG
  OptRNGType := TRNGType(ReadInteger(SECTION_RNG, KEY_USE_RNG, DFLT_KEY_USE_RNG));


  // Generate
  OptPasswordLength         := ReadInteger(SECTION_GENERATE, KEY_PASSWORD_LENGTH,   DFLT_KEY_PASSWORD_LENGTH);
  OptPasswordStrength       := ReadInteger(SECTION_GENERATE, KEY_PASSWORD_STRENGTH, DFLT_KEY_PASSWORD_STRENGTH);
  OptTargetPasswordByLength := ReadBool(SECTION_GENERATE, KEY_TARGET_PASSWORD_BASIS, DFLT_KEY_TARGET_PASSWORD_BASIS);

  Result := TRUE;
end;



function TPasswordGenIniFile.WriteIniFileOptions() : boolean;
var
  thisMajorVersion   : integer;
  thisMinorVersion   : integer;
  thisRevisionVersion: integer;
  thisBuildVersion   : integer;

  strVersion: string;
begin
  SDUGetVersionInfo('', thisMajorVersion, thisMinorVersion, thisRevisionVersion, thisBuildVersion);
  strVersion := SDUGetVersionInfoString('');
  if (APP_BETA_BUILD>-1) then
    begin
    strVersion := strVersion + ' BETA ' + inttostr(APP_BETA_BUILD);
    end;

  WriteString(SECTION_PASSWORDGEN, KEY_VERSION, strVersion);
  WriteInteger(SECTION_PASSWORDGEN, KEY_MAJOR_VERSION,   thisMajorVersion);
  WriteInteger(SECTION_PASSWORDGEN, KEY_MINOR_VERSION,   thisMinorVersion);
  WriteInteger(SECTION_PASSWORDGEN, KEY_RELEASE_VERSION, thisRevisionVersion);
  WriteInteger(SECTION_PASSWORDGEN, KEY_BUILD_VERSION,   thisBuildVersion);
  WriteInteger(SECTION_PASSWORDGEN, KEY_BETA_VERSION,    APP_BETA_BUILD);



  // Options
  WriteBool(SECTION_OPTIONS, KEY_COPY_PASSWORDS_TO_CLIPBOARD, OptCopyPasswordsToClipboard);
  WriteBool(SECTION_OPTIONS, KEY_WARN_PASSWORDS_COPIED_TO_CLIPBOARD, OptWarnPasswordsCopiedToClipboard);
  WriteBool(SECTION_OPTIONS, KEY_NO_START_END_IN_SPACE, OptNoStartEndInSpace);


  // Character sets available
  WriteBool(SECTION_CHARACTERSETS, KEY_USE_UPPERCASE,   OptUseUppercase);
  WriteBool(SECTION_CHARACTERSETS, KEY_USE_LOWERCASE,   OptUseLowercase);
  WriteBool(SECTION_CHARACTERSETS, KEY_USE_NUMBERS,     OptUseNumbers);
  WriteBool(SECTION_CHARACTERSETS, KEY_USE_PUNCTUATION, OptUsePunctuation);
  WriteBool(SECTION_CHARACTERSETS, KEY_USE_SPACE,       OptUseSpace);

  WriteString(SECTION_CHARACTERSETS, KEY_PUNCTUATION_SET, OptPunctuationSet);

  WriteBool(SECTION_CHARACTERSETS, KEY_FORCE_CHAR_FROM_SELECTED_CHARSETS, OptForceCharFromEachSelectedCharSet);
  WriteBool(SECTION_CHARACTERSETS, KEY_FORCE_CHAR_FROM_SELECTED_CHARSETS_NOT_SPACE, OptForceCharFromEachSelectedCharSetExceptSpace);


  // RNG
  WriteInteger(SECTION_RNG, KEY_USE_RNG, ord(OptRNGType));


  // Generate
  WriteInteger(SECTION_GENERATE, KEY_PASSWORD_LENGTH,   OptPasswordLength);
  WriteInteger(SECTION_GENERATE, KEY_PASSWORD_STRENGTH, OptPasswordStrength);
  WriteBool(SECTION_GENERATE, KEY_TARGET_PASSWORD_BASIS, OptTargetPasswordByLength);

  Result := TRUE;
end;


procedure TPasswordGenIniFile.CheckVersionID();
var
  oldVersionStr: string;
  upgradeMessage: string;

  majorVersion   : integer; // Only used to build up oldVersionStr
  minorVersion   : integer; // Only used to build up oldVersionStr
//  revisionVersion: integer; // Only used to build up oldVersionStr
//  buildVersion   : integer; // Only used to build up oldVersionStr

  thisMajorVersion   : integer;
  thisMinorVersion   : integer;
  thisRevisionVersion: integer;
  thisBuildVersion   : integer;
  strThisVersion     : string;

  betaVersion: integer;
begin
  oldVersionStr    := ReadString(SECTION_PASSWORDGEN, KEY_VERSION, '-1.0');
  majorVersion     := ReadInteger(SECTION_PASSWORDGEN, KEY_MAJOR_VERSION, -1);
  minorVersion     := ReadInteger(SECTION_PASSWORDGEN, KEY_MINOR_VERSION, -1);
//  revisionVersion  := ReadInteger(SECTION_PASSWORDGEN, KEY_RELEASE_VERSION, -1);
//  buildVersion     := ReadInteger(SECTION_PASSWORDGEN, KEY_BUILD_VERSION, -1);
  betaVersion      := ReadInteger(SECTION_PASSWORDGEN, KEY_BETA_VERSION, -1);

  SDUGetVersionInfo('', thisMajorVersion, thisMinorVersion, thisRevisionVersion, thisBuildVersion);
  strThisVersion := Format('%d.%.2d.%.2d.%.4d', [thisMajorVersion, thisMinorVersion, thisRevisionVersion, thisBuildVersion]);
  if (APP_BETA_BUILD>-1) then
    begin
    strThisVersion := strThisVersion + ' BETA ' + inttostr(APP_BETA_BUILD);
    end;

  NewVersion := FALSE;

  if oldVersionStr='-1.0' then
    begin
    // Any message to be displayed the 1st time the user runs the application
    // to be placed here
    NewVersion := TRUE;
    end
  else if (majorVersion<thisMajorVersion) OR
          ((majorVersion=thisMajorVersion) AND (minorVersion<thisMinorVersion)) OR
          ((betaVersion<>-1) AND (not(APP_BETA_BUILD>-1))) OR
          ((betaVersion<>-1) AND (betaVersion<APP_BETA_BUILD)) then
    begin
    if betaVersion>-1 then
      begin
      oldVersionStr := oldVersionStr + ' BETA ' + inttostr(betaVersion);
      end;

    upgradeMessage := 'It appears as though your "PasswordGen.ini" file dates back to v'+oldVersionStr;
    upgradeMessage := upgradeMessage+CRLF+CRLF+'You may find that some of your options mysteriously revert back to their defaults when running this version (v'+strThisVersion+').';
    upgradeMessage := upgradeMessage+CRLF+CRLF+'This is "normal", and due to two things:';
    upgradeMessage := upgradeMessage+CRLF+'1) the format of the ini file has changed to make it more logical and to take into account improvements in the software';
    upgradeMessage := upgradeMessage+CRLF+'2) lazyness on the part of the author (it wouldn''t be difficult to write a converter, but it''s just too much hassle)';
    upgradeMessage := upgradeMessage+CRLF+CRLF+'To prevent this dialog from reappearing, either delete your existing SecureTrayUtil.ini file (located in your Windows directory) or goto any of the options screens, change an option, change it back, and click "OK" or "Apply"';
    upgradeMessage := upgradeMessage+CRLF+CRLF+'It is recommended that you check your option settings to ensure correct operation.';

    messagedlg(upgradeMessage, mtWarning, [mbOK], 0);

    NewVersion := TRUE;
    end;

end;


END.


