unit Main;
// Description: PasswordGen - A secure password generating program
// By Sarah Dean
// Email: sdean12@softhome.net
// WWW:   http://www.SDean12.org/
//
// -----------------------------------------------------------------------------



interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Spin, ExtCtrls, ComCtrls, Menus,
  INIFile,
  SDLargeDivision;


type
  TfrmPasswordGen = class(TForm)
    edPassword: TEdit;
    Label1: TLabel;
    GroupBox2: TGroupBox;
    seLength: TSpinEdit;
    rbSizeByLength: TRadioButton;
    Label3: TLabel;
    reReport: TRichEdit;
    Label4: TLabel;
    pbGeneratePassword: TButton;
    Button1: TButton;
    seStrength: TSpinEdit;
    Label2: TLabel;
    rbSizeByStrength: TRadioButton;
    pbOptions: TButton;
    pbAbout: TButton;
    pbPrint: TButton;
    PrintDialog1: TPrintDialog;
    rePrint: TRichEdit;
    Label5: TLabel;
    Label6: TLabel;
    edEstPassEntropy: TEdit;
    pbTest: TButton;
    procedure pbGeneratePasswordClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure seLengthStrengthChange(Sender: TObject);
    procedure rbSizeByLengthStrength(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure pbOptionsClick(Sender: TObject);
    procedure pbAboutClick(Sender: TObject);
    procedure pbPrintClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure pbTestClick(Sender: TObject);
  private
    INIFile: TPasswordGenIniFile;

    // Flag to indicate that the system is just starting up
    Initializing: boolean;

    function DetermineDrawSet(var drawset: string): boolean;
    function GeneratePasswordStrength(var drawSet: string; strengthInBits: integer; var password: string): boolean;
    function GeneratePasswordLength(var drawSet: string; passwordLength: integer; var password: string): boolean;
    function CheckPasswordValid(var password: string): boolean;

    function SetHugeIntToRandom(strengthInBits: integer; var randData: TSDLargeNum): boolean;
    procedure ConvertRandomToPassword(var drawSet: string; randData: TSDLargeNum; var password: string);

    procedure OverwriteString(var data: string);

    procedure DisplayPassword(var drawSet: string; var password: string);

    function  CalcMaxPasswordEntropy(): integer;
    function  CalcDrawsetBitsPerChar(var drawset: string): double;

    // Given the specified drawset and password, estimate the amount of entropy it has
    function  CalcPasswordStrength(var drawset: string; var password: string): double;

    procedure SetupControls();
  public
    { Public declarations }
  end;

var
  frmPasswordGen: TfrmPasswordGen;

implementation

{$R *.DFM}


uses
  Math, // required for "log2" function
  Clipbrd, // required for clipboard functions
  CharConsts,
  AppGlobalConsts,
  Options,
  RandomGenerator,
  About,
  wcrypt2;  // Required for the Microsoft RNG; part of the JEDI project

const
  WARN_AFTER_GENERATE_ATTEMPTS = 10;

  CRLF = #10+#13;
  TAB  = #09;


procedure TfrmPasswordGen.pbGeneratePasswordClick(Sender: TObject);
var
  password: string;
  passwordOK: boolean;
  drawset: string;
  msgDlgReply: word;
  attempts: integer;
begin
  edPassword.text := 'Generating, please wait...';
  edEstPassEntropy.text := '';
  reReport.lines.clear();
  SetupControls();
  Application.ProcessMessages();


  DetermineDrawset(drawset);

  // This restriction is included because:
  //   1) We use "char" - an 8 bit datatype - to handle password characters
  //   2) The large integer functions operate on an array of *bytes* - 8 bit
  //      numbers
  if (length(drawset)>256) then
    begin
    MessageDlg(
               'This version of the software can only operate with a character set of '+CRLF+
               'less 256 characters or less.'+CRLF+
               CRLF+
               'Please change your character set options to reduce the number of '+CRLF+
               'possible characters.',
               mtError,
               [mbOK],
               0
               );
    exit;
    end;

  attempts := 0;

  // Generate passwords, until we have a valid one
  passwordOK := FALSE;
  while not(passwordOK) do
    begin
    if rbSizeByLength.checked then
      begin
      passwordOK := GeneratePasswordLength(drawSet, seLength.value, password);
      end
    else
      begin
      passwordOK := GeneratePasswordStrength(drawSet, seStrength.value, password);
      end;

    if not(passwordOK) then
      begin
      // User cancelled, or other error preventing the system from generating
      // a random password
      break;
      end;

    passwordOK := CheckPasswordValid(password);

    if not(passwordOK) then
      begin
      if (INIFile.OptRNGType = RNGMouse) then
        begin
        msgDlgReply := MessageDlg('The password generated is not valid'+CRLF+
                                  'according to the options you have selected.'+CRLF+
                                  CRLF+
                                  'Try again?',
                                  mtWarning,
                                  [mbYes, mbNo],
                                  0);
        if (msgDlgReply=mrNo) then
          begin
          passwordOK := TRUE;
          end;
        end;

      end;


    if not(passwordOK) then
      begin
      inc(attempts);
      if (attempts > WARN_AFTER_GENERATE_ATTEMPTS) then
        begin
        msgDlgReply := MessageDlg('After '+inttostr(WARN_AFTER_GENERATE_ATTEMPTS)+' attempts to generate a valid password according to the options selected, no valid password has been generated.'+CRLF+
                                  CRLF+
                                  'You may wish to revise your options to remove one or more constraints when checking if a password generated is valid'+CRLF+
                                  CRLF+
                                  'Do you wish to try another '+inttostr(WARN_AFTER_GENERATE_ATTEMPTS)+' times to genreate a valid password?',
                                  mtWarning,
                                  [mbYes, mbNo],
                                  0);
        if (msgDlgReply=mrNo) then
          begin
          passwordOK := TRUE;
          end
        else
          begin
          attempts := 0;
          end;

        end;

      end;

    end;


  DisplayPassword(drawset, password);


  if (password<>'') then
    begin
    if INIFile.OptCopyPasswordsToClipboard then
      begin
      Clipboard.AsText := password;
      if INIFile.OptWarnPasswordsCopiedToClipboard then
        begin
        MessageDlg(
                   'The password generated has been copied to the clipboard.',
                   mtInformation,
                   [mbOK],
                   0
                  );
        end;

      end;
    end;


  OverwriteString(password);
  OverwriteString(drawset);

end;


function TfrmPasswordGen.GeneratePasswordLength(var drawSet: string; passwordLength: integer; var password: string): boolean;
var
  bitsRequired: integer;
  allOK: boolean;
  bitsPerChar: double;
begin
  allOK := TRUE;

  password := '';


  bitsPerChar := CalcDrawsetBitsPerChar(drawset);


  // Calculate the number of bits of entropy that a password of length
  // "length", using the chars in "drawSet" must have
  // i.e. <number of chars> to the power of <length>
  // We carry out this operation
  bitsRequired := trunc(bitsPerChar * passwordLength);
  // Round up...
  if (bitsRequired < (bitsPerChar * passwordLength)) then
    begin
    bitsRequired := bitsRequired + 1;
    end;

  if (allOK) then
    begin
    allOK := GeneratePasswordStrength(drawSet, bitsRequired, password)
    end;

  // If the password generated was longer than required (e.g. because the RNG
  // only generates random data 8 bits at a time), truncate the password to
  // the required length
  // We delete characters from the START of the password; the char with the MSB
  // may not have as much entropy as the other chars
  // If it still has too many chars, we delete from the END of the password
  // e.g. if the alphabet has only 3 chars; A, B and C - and the amount of
  // entropy is 3 bits, then:
  // 000 -> AAA
  // 001 -> AAB
  // 010 -> AAC
  // 011 -> ABA
  // 100 -> ABB
  // 101 -> ABC
  // 110 -> BAA
  // 111 -> BAB
  //   - note that the MSB is more likely to be "A". If we've generated too
  //     many chars for the password, then we must have generated a password
  //     with more entropy than required; the password needs to be trimmed down
  //     to the correct size
  //   - also note that the LSB is most likely to be A/B. This is OK since the
  //     amount of entropy in the password doesn't change
  while (length(password) > passwordLength) do
    begin
    Delete(password, 1, 1);

    // If we still have too man chars, trim from the other end
    if (length(password) > passwordLength) then
      begin
      Delete(password, length(password), 1);
      end;
    end;

  Result := allOK;
end;

function TfrmPasswordGen.GeneratePasswordStrength(var drawSet: string; strengthInBits: integer; var password: string): boolean;
var
  randData: TSDLargeNum;
  allOK: boolean;
  tmpInt: integer;
  generateRandomBits: integer;
begin
  password := '';

  // Generate the random data
  // (Initialize first)
  SD_Large_zero(randData);

  // Round "strengthInBits" up to the nearest multiple of 8
  generateRandomBits := strengthInBits;
  tmpInt := generateRandomBits mod 8;
  if tmpInt<>0 then
    begin
    generateRandomBits := generateRandomBits + (8 - tmpInt);
    end;

  allOK := SetHugeIntToRandom(generateRandomBits, randData);

  // Convert the random data into base-n, where "n" is the length of the
  // drawSet; thereby converting it into a password
  if (allOK) then
    begin
    ConvertRandomToPassword(drawSet, randData, password);

    // At this point, the password generated may not reflect the randData
    // accurately (e.g. if the password strength was 128, but randData was
    // randomly set to int(1), then the password will only have 1 char.
    // The password should then be padded out with the base-n "0" char
    while (CalcPasswordStrength(drawset, password) < strengthInBits) do
      begin
      // Note: +1 because the drawSet string indexes from 1
      password := drawSet[1] + password;
      end;

    end;




  Result := allOK;
end;


procedure TfrmPasswordGen.ConvertRandomToPassword(var drawSet: string; randData: TSDLargeNum; var password: string);
var
  baseN: integer;
  divValue: TSDLargeNum;
  modValue: byte;
begin
  baseN := length(drawSet);

  modValue := 0;
  while (SD_Large_compare(randData, baseN) >= 0) do
    begin
    SD_Large_divide(randData, baseN, divValue, modValue);

    // Convert the remaining value in randData to it's integer representation
    // Note: +1 because the drawSet string indexes from 1
    password := drawSet[modValue+1] + password;

    SD_Large_copy(divValue, randData);
    end;


  // At this point, the value of randData is greater than baseN; we do a
  // divide to transfer randData into modValue
  // Convert the remaining value in randData to it's integer representation
  SD_Large_divide(randData, baseN, divValue, modValue);

  // Note: +1 because the drawSet string indexes from 1
  password := drawSet[modValue+1] + password;

end;



{
procedure TfrmPasswordGen.ConvertRandomToPassword(var drawSet: string; randData: HugeInt; var password: string);
var
  hiBaseN: HugeInt;
  baseN: integer;
  tmpStr: string;
  divValue: HugeInt;
  modValue: HugeInt;
  charIdx: integer;
  i: integer;
  s: string;
  a: HugeInt;
  z: HugeInt;
begin
  baseN := length(drawSet);
  Integer2HugeInt(baseN, hiBaseN);


  while (HugeInt_Comp(randData, hiBaseN) >= 0) do
    begin
    reReport.lines.add('');
    s:='';
    for i:=10 downto 1 do
      begin
      s := s + inttohex(randData[i-1], 2);
      end;
   reReport.lines.add('randata  = '+s);


//    HugeInt_Div(randData, hiBaseN, divValue);
a := randData;
    HugeInt_Div(a, hiBaseN, divValue);




    s:='';
    for i:=10 downto 1 do
      begin
      s := s + inttohex(divValue[i-1], 2);
      end;
   reReport.lines.add('divValue = '+s);

//    HugeInt_Mod(randData, hiBaseN, modValue);
a := randData;
    HugeInt_Mod(a, hiBaseN, modValue);

    Integer2HugeInt(0, z);
    HugeInt_Mul(divValue, hiBaseN, z);
    HugeInt_Add(z, modValue, z);
    s:='';
    for i:=10 downto 1 do
      begin
      s := s + inttohex(z[i-1], 2);
      end;
   reReport.lines.add('z             = '+s);

    // Convert the remaining value in randData to it's integer representation
    // Note: This will *never* set charIdx to a value which it can't store
    //        - unless the number of chars in drawSet are greater than the
    //          max number it can store! (i.e. this isn't ever going to happen)
    HugeInt2String(modValue, tmpStr);
    charIdx := strtoint(tmpStr);

    // Note: +1 because the drawSet string indexes from 1
    password := drawSet[charIdx+1] + password;

    HugeInt_Copy(divValue, randData);
//    randData := divValue;
    end;


  // Convert the remaining value in randData to it's integer representation
  // Note: This will *never* set charIdx to a value which it can't store
  //        - unless the number of chars in drawSet are greater than the
  //          max number it can store! (i.e. this isn't ever going to happen)
  HugeInt2String(randData, tmpStr);
  charIdx := strtoint(tmpStr);

  // Note: +1 because the drawSet string indexes from 1
  password := drawSet[charIdx+1] + password;

end;
}



{
// Old version - uses recursion and can cause stack overflows when generating
// passwords with large amounts of entropy
procedure TfrmPasswordGen.ConvertRandomToPassword(var drawSet: string; randData: HugeInt; var password: string);
var
  hiBaseN: HugeInt;
  baseN: integer;
  tmpStr: string;
  divValue: HugeInt;
  modValue: HugeInt;
  charIdx: integer;
begin
  baseN := length(drawSet);
  Integer2HugeInt(baseN, hiBaseN);

  if (HugeInt_Comp(randData, hiBaseN) = -1) then
    begin
    // Convert the remaining value in randData to it's integer representation
    // Note: This will *never* set charIdx to a value which it can't store
    //        - unless the number of chars in drawSet are greater than the
    //          max number it can store! (i.e. this isn't ever going to happen)
    HugeInt2String(randData, tmpStr);
    charIdx := strtoint(tmpStr);

    // Note: +1 because the drawSet string indexes from 1
    password := password + drawSet[charIdx+1];
    end
  else
    begin
    HugeInt_Div(randData, hiBaseN, divValue);
    HugeInt_Mod(randData, hiBaseN, modValue);

    ConvertRandomToPassword(drawSet, divValue, password);

    // Convert the remaining value in randData to it's integer representation
    // Note: This will *never* set charIdx to a value which it can't store
    //        - unless the number of chars in drawSet are greater than the
    //          max number it can store! (i.e. this isn't ever going to happen)
    HugeInt2String(modValue, tmpStr);
    charIdx := strtoint(tmpStr);

    // Note: +1 because the drawSet string indexes from 1
    password := password + drawSet[charIdx+1];
    end;

end;
}


function TfrmPasswordGen.DetermineDrawSet(var drawset: string): boolean;
var
  allOK: boolean;
begin
  allOK:= FALSE;
  drawset := '';

  if INIFile.OptUseUppercase then
    begin
    drawset := drawset + CHARSET_UPPERCASE;
    end;

  if INIFile.OptUseLowercase then
    begin
    drawset := drawset + CHARSET_LOWERCASE;
    end;

  if INIFile.OptUseNumbers then
    begin
    drawset := drawset + CHARSET_NUMBERS;
    end;

  if INIFile.OptUsePunctuation then
    begin
    drawset := drawset + INIFile.OptPunctuationSet;
    end;

  if INIFile.OptUseSpace then
    begin
    drawset := drawset + CHARSET_WHITESPACE;
    end;

  if length(drawset)<=0 then
    begin
    showmessage('Please select one or more chacter sets from which passwords should be generated');
    end
  else if length(drawset)=1 then
    begin
    showmessage('A password generated using only a single character will not be very secure...'+CRLF+CRLF+'Please select more character sets from which passwords should be generated');
    end
  else
    begin
    allOK := TRUE;
    end;

  Result := allOK;

end;


procedure TfrmPasswordGen.FormCreate(Sender: TObject);
begin
  Initializing := TRUE;

  INIFile:= TPasswordGenIniFile.Create(INI_FILENAME);
  INIFile.ReadIniFile();


  // Password size
  seLength.value   := INIFile.OptPasswordLength;
  seStrength.value := INIFile.OptPasswordStrength;
  if INIFile.OptTargetPasswordByLength then
    begin
    rbSizeByLength.checked := TRUE;
    end
  else
    begin
    rbSizeByStrength.checked := TRUE;
    end;


  Initializing := FALSE;

end;

procedure TfrmPasswordGen.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  INIFile.WriteIniFile();

  INIFile.Free();

end;

procedure TfrmPasswordGen.seLengthStrengthChange(Sender: TObject);
begin
  if not(Initializing) then
    begin
    INIFile.OptPasswordLength   := seLength.value;
    INIFile.OptPasswordStrength := seStrength.value;
    end;

end;

procedure TfrmPasswordGen.rbSizeByLengthStrength(Sender: TObject);
begin
  if not(Initializing) then
    begin
    INIFile.OptTargetPasswordByLength := rbSizeByLength.checked;
    end;

end;


function TfrmPasswordGen.CheckPasswordValid(var password: string): boolean;
var
  retVal: boolean;
  i: integer;
  flagUppercase  : boolean;
  flagLowercase  : boolean;
  flagNumbers    : boolean;
  flagPunctuation: boolean;
  flagSpace      : boolean;
begin
  retVal := TRUE;

  // Check to see if the user specified that all charsets must be represented
  // in the password
  if INIFile.OptForceCharFromEachSelectedCharSet OR
     INIFile.OptForceCharFromEachSelectedCharSetExceptSpace then
    begin
    flagUppercase  := FALSE;
    flagLowercase  := FALSE;
    flagNumbers    := FALSE;
    flagPunctuation:= FALSE;
    flagSpace      := FALSE;

    for i:=1 to length(password) do
      begin
      if pos(password[i], CHARSET_UPPERCASE)>0 then
        begin
        flagUppercase := TRUE;
        end;

      if pos(password[i], CHARSET_LOWERCASE)>0 then
        begin
        flagLowercase := TRUE;
        end;

      if pos(password[i], CHARSET_NUMBERS)>0 then
        begin
        flagNumbers := TRUE;
        end;

      if pos(password[i], INIFile.OptPunctuationSet)>0 then
        begin
        flagPunctuation := TRUE;
        end;

      if pos(password[i], CHARSET_WHITESPACE)>0 then
        begin
        flagSpace := TRUE;
        end;

      end;

    if INIFile.OptForceCharFromEachSelectedCharSet then
      begin
      if (INIFile.OptUseUppercase   and not(flagUppercase))   OR
         (INIFile.OptUseLowercase   and not(flagLowercase))   OR
         (INIFile.OptUseNumbers     and not(flagNumbers))     OR
         (INIFile.OptUsePunctuation and not(flagPunctuation)) OR
         (INIFile.OptUseSpace       and not(flagSpace))       then
        begin
        retVal := FALSE;
        end;
      end
    else if INIFile.OptForceCharFromEachSelectedCharSetExceptSpace then
      begin
      if (INIFile.OptUseUppercase   and not(flagUppercase))   OR
         (INIFile.OptUseLowercase   and not(flagLowercase))   OR
         (INIFile.OptUseNumbers     and not(flagNumbers))     OR
         (INIFile.OptUsePunctuation and not(flagPunctuation)) then
        begin
        retVal := FALSE;
        end;
      end;

    end;


    // Check to see if user pemits starting/ending with space
    if ( INIFile.OptNoStartEndInSpace AND
         ( (password[1]=' ') OR
           (password[length(password)]=' ') ) ) then
      begin
      retVal := FALSE;
      end;

  Result := retVal;
end;


procedure TfrmPasswordGen.Button1Click(Sender: TObject);
begin
  Close();

end;

procedure TfrmPasswordGen.pbOptionsClick(Sender: TObject);
var
  frm: TfrmOptions;
begin
  frm:= TfrmOptions.Create(nil);
  try
    frm.INIFile := INIFile;
    frm.ShowModal();
    SetupControls();
  finally
    frm.Free();
  end;

end;


procedure TfrmPasswordGen.pbAboutClick(Sender: TObject);
var
  frm: TfrmAbout;
begin
  frm:= TfrmAbout.Create(nil);
  try
    frm.ShowModal();
  finally
    frm.Free();
  end;

end;


// Populate randData with strengthInBits bits of random data
// !!! WARNING !!!
// "strengthInBits" MUST BE a multiple of 8
function TfrmPasswordGen.SetHugeIntToRandom(strengthInBits: integer; var randData: TSDLargeNum): boolean;
var
  rngFrm: TRandomGenerator_F;
  allOK: boolean;
  hProvider: HCRYPTPROV;
//  i: integer;
begin
  allOK := FALSE;
  
  if (INIFile.OptRNGType = RNGMouse) then
    begin
    rngFrm:= TRandomGenerator_F.Create(nil);
    try
      rngFrm.RequiredBits := strengthInBits;
      rngFrm.BitsAlreadyGot := 0;
      if (rngFrm.ShowModal() = mrOK) then
        begin
        rngFrm.GetBitsCollected(strengthInBits, randData);
        allOK := TRUE;
        end;

    finally
      rngFrm.Free();
    end;

    end
  else if (INIFile.OptRNGType = RNGMicrosoft) then
    begin
    if (not(CryptAcquireContext(@hProvider, nil, nil, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))) then
      begin
      MessageDlg('Unable to obtain RNG context from Microsoft RNG.', mtError, [mbOK], 0);
      end
    else
      begin
      if (CryptGenRandom(hProvider, (strengthInBits div 8), @randData[0])) then
        begin
        allOK:= TRUE;
        end
      else
        begin
        MessageDlg('Unable to obtain random data from Microsoft RNG.', mtError, [mbOK], 0);
        end;


      // Debug: Set all bits to 1
//      for i:=1 to ((strengthInBits div 8)-1) do
////      for i:=1 to 9 do
//        begin
//        randData[i-1] := $ff;
//        end;


      // When this next one is tried with the HugeInt library, it fails to
      // generate the correct output with a 17 char-long drawset
//      SD_Large_zero(randData);
//      randData[0] := $0f;
//      randData[1] := $2E;
//      randData[2] := $4C;
//      randData[3] := $6A;
//      randData[4] := $88;
//      randData[5] := $A6;
//      randData[6] := $C4;
//      randData[7] := $E2;


      CryptReleaseContext(hProvider, 0);
      end;

    end
  else
    begin
    showmessage('Unknown RNG?!!');
    end;


  Result := allOK;
end;


procedure TfrmPasswordGen.OverwriteString(var data: string);
begin
  // +5 just for the hell of it; overwrite past existing string length
  data := StringOfChar(#0, length(data) + 5);

end;

procedure TfrmPasswordGen.DisplayPassword(var drawSet: string; var password: string);
var
  charToSpoken: TCharToSpoken;
  spokenChar: string;
  i: integer;
  passwordStrength: double;
begin
  edPassword.text := password;

  if (password='') then
    begin
    reReport.lines.Clear();
    end
  else
    begin
    if (password[1]=' ') then
      begin
      MessageDlg('WARNING: Password starts with whitespace', mtWarning, [mbOK], 0);
      end;

    if (password[length(password)]=' ') then
      begin
      MessageDlg('WARNING: Password ends with whitespace', mtWarning, [mbOK], 0);
      end;

    charToSpoken:= TCharToSpoken.Create();
    try
      for i:=1 to length(password) do
        begin
        spokenChar := charToSpoken.GetSpokenChar(password[i]);
        reReport.lines.add(spokenChar);
        end;

    finally
      charToSpoken.Free();
    end;

    end;  // if (password='') then

  passwordStrength := CalcPasswordStrength(drawset, password);
  // We only display to 4 decimal places. That's more than enough anyway...
  edEstPassEntropy.Text := FormatFloat('##########.####', passwordStrength);


  SetupControls();

end;

procedure TfrmPasswordGen.pbPrintClick(Sender: TObject);
begin
  if PrintDialog1.Execute() then
    begin
    rePrint.lines.Clear();
    rePrint.WordWrap := FALSE;

    rePrint.lines.Add('Password generated: '+edPassword.text);

    if (edPassword.text[1]=' ') then
      begin
      rePrint.lines.Add('WARNING: Password starts with whitespace');
      end;

    if (edPassword.text[length(edPassword.text)]=' ') then
      begin
      rePrint.lines.Add('WARNING: Password ends with whitespace');
      end;

    // Change the font, etc for the password
    rePrint.SelStart:= Length(rePrint.lines[0]) - length(edPassword.text);
    rePrint.SelLength := length(edPassword.text);
//    rePrint.SelAttributes.color := clred;
    rePrint.SelAttributes.Name := 'Courier';
    rePrint.SelAttributes.Pitch := fpFixed;
    rePrint.SelAttributes.Style := [fsBold];

    rePrint.lines.Add('');
    rePrint.lines.Add('Spoken representation:');
    rePrint.Paragraph.FirstIndent := 10;
    rePrint.lines.AddStrings(reReport.lines);

    rePrint.Paragraph.FirstIndent := 0;
    rePrint.lines.Add('');
    rePrint.lines.Add('Password entropy: '+edEstPassEntropy.text+' bits.');

    rePrint.Print('Password Generated...');

    rePrint.Text := StringOfChar('Z', length(rePrint.Text));
    end;

end;

procedure TfrmPasswordGen.FormShow(Sender: TObject);
begin
  edPassword.Text := '';
  edEstPassEntropy.Text := '';

  SetupControls();

end;


// Calculate the max amount of password entropy which can be achieved with the
// current HugeInt data type
function TfrmPasswordGen.CalcMaxPasswordEntropy(): integer;
begin
  // HugeInt stores numbers as "n" 8-bit bytes, including the MSB sign bit
  // Therefore, we'll ignore the last byte (as it includes the sign bit, which
  // we really want nothing to do with), and multiply by 8, giving:
  Result := ((sizeof(TSDLargeNum) - 1) * 8)

end;

// Calculate the amount of bits of entropy each character in drawSet
// represents
function TfrmPasswordGen.CalcDrawsetBitsPerChar(var drawset: string): double;
begin
  // <bits per char> = log2(<count of chars in drawSet>)
  Result := (log2(length(drawSet)));

end;


procedure TfrmPasswordGen.SetupControls();
var
  drawset: string;
  maxBits: integer;
  maxLength: integer;
  maxLengthAccurate: double;
  bitsPerChar: double;
begin
  pbPrint.enabled := (reReport.text <> '');

  DetermineDrawset(drawset);

  // Calc the max length of a password
  maxBits := CalcMaxPasswordEntropy();
  bitsPerChar := CalcDrawsetBitsPerChar(drawset);

  maxLengthAccurate := maxBits / bitsPerChar;
  // Round down to the nearest char...
  maxLength := trunc(maxLengthAccurate);

  OverwriteString(drawset);

  seStrength.MaxValue := maxBits;
  seLength.MaxValue := maxLength;

end;


// Given the specified drawset and password, estimate the amount of entropy it has
function TfrmPasswordGen.CalcPasswordStrength(var drawset: string; var password: string): double;
var
  bitsPerChar: double;
begin
  bitsPerChar := CalcDrawsetBitsPerChar(drawset);
  Result := bitsPerChar * Length(password);

end;


// This is a simple test: Show the quotient and remainder of:
//   E2C4A6886A4C2E0F (hex) / 11 (hex)
// The correct result is:
//   Quotient: D56DC9E9CD74E00 (hex)
//   Remainder: F (hex)
// This test exists because the "HugeInt" library (part of the UDDF) previously
// used returned the wrong values!!!
procedure TfrmPasswordGen.pbTestClick(Sender: TObject);
var
  start: TSDLargeNum;
  outDiv: TSDLargeNum;
  outMod: byte;
  s: string;
  i: integer;
begin
  SD_Large_zero(start);

  start[0] := $0f;
  start[1] := $2E;
  start[2] := $4C;
  start[3] := $6A;
  start[4] := $88;
  start[5] := $A6;
  start[6] := $C4;
  start[7] := $E2;

  SD_Large_divide(start, 17, outDiv, outMod);

  s:='';
  for i:=10 downto 1 do
    begin
    s := s + inttohex(start[i-1], 2);
    end;
  reReport.lines.add('The number: '+s+' (hex)');

  reReport.lines.add('Divided by 11 (hex; 17 in decimal), gives:');

  s:='';
  for i:=10 downto 1 do
    begin
    s := s + inttohex(outDiv[i-1], 2);
    end;
  reReport.lines.add('Quotient: '+s+' (hex)');
  reReport.lines.add('Remainder: '+inttohex(outMod, 1)+' (hex)');
  reReport.lines.add('');

end;

END.


